// Copyright (C) 1999-2001 Open Source Telecom Corporation.
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// 
// As a special exception to the GNU General Public License, permission is 
// granted for additional uses of the text contained in its release 
// of Common C++.
// 
// The exception is that, if you link the Common C++ library with other
// files to produce an executable, this does not by itself cause the
// resulting executable to be covered by the GNU General Public License.
// Your use of that executable is in no way restricted on account of
// linking the Common C++ library code into it.
//
// This exception does not however invalidate any other reasons why
// the executable file might be covered by the GNU General Public License.
// 
// This exception applies only to the code released under the 
// name Common C++.  If you copy code from other releases into a copy of
// Common C++, as the General Public License permits, the exception does
// not apply to the code that you add in this way.  To avoid misleading
// anyone as to the status of such modified files, you must delete
// this exception notice from them.
// 
// If you write modifications of your own for Common C++, it is your choice
// whether to permit this exception to apply to your modifications.
// If you do not wish that, delete this exception notice.  

#include "config.h"
#ifndef	WIN32
#include "macros.h"
#else
#include "export.h"
#endif
#include "thread.h"
#include "socket.h"
#include <fcntl.h>
#include <errno.h>
#include <cstdlib>
#ifndef	WIN32
#include <netinet/tcp.h>
#endif
#include <strstream>

#ifdef	WIN32
#include <io.h>
#endif

#ifndef	WIN32
# ifndef  O_NONBLOCK
#  define O_NONBLOCK	O_NDELAY
# endif
# ifdef	IPPROTO_IP
#  ifndef  SOL_IP
#   define SOL_IP	IPPROTO_IP
#  endif // !SOL_IP
# endif	 // IPPROTO_IP
#endif	 // !WIN32

using namespace std;

Socket::Socket(int domain, int type, int protocol)
{
	setSocket();
	so = socket(domain, type, protocol);
	if(so == INVALID_SOCKET)
	{
		Error(SOCKET_CREATE_FAILED);
		return;
	}
	state = SOCKET_AVAILABLE;
}

Socket::Socket(SOCKET fd)
{
	setSocket();
	if (fd == INVALID_SOCKET)
	{
		Error(SOCKET_CREATE_FAILED);
		return;
	}
	so = fd;
	state = SOCKET_AVAILABLE;
}

Socket::Socket(const Socket &orig)
{
	setSocket();
#ifdef WIN32
	so = orig.so;
#else
	so = dup(orig.so);
#endif
	if(so == INVALID_SOCKET)
		Error(SOCKET_COPY_FAILED);
	state = orig.state;
}

void Socket::setSocket(void)
{
	flags.thrown    = false;
	flags.broadcast = false;
	flags.route     = true;
	flags.keepalive = false;
	flags.loopback  = true;
	flags.multicast = false;
	flags.linger	= false;
	flags.ttl	= 1;
	errid           = SOCKET_SUCCESS;
	errstr          = NULL;
	state           = SOCKET_INITIAL;
	so              = -1;
}

ssize_t Socket::Readline(char *str, size_t request, timeout_t timeout)
{
	bool crlf = false;
	bool nl = false;
	size_t nleft = request;
	ssize_t nstat;
//	char *orig;
	int c;

	if(request < 1)
		return 0;

	str[0] = 0;

	while(nleft && !nl)
	{
		if(timeout)
		{
			if(!isPending(SOCKET_PENDING_INPUT, timeout))
			{
				Error(SOCKET_TIMEOUT_ERROR);
				return -1;
			}
		}
		nstat = ::recv(so, str, nleft, MSG_PEEK);
		if(nstat <= 0)
		{
			Error(SOCKET_INPUT_ERROR);
			return -1;
		}

		c = 0;
		while(c < nstat && !nl)
		{
			if(str[c] == '\r' && c < nstat - 1)
			{
				if(str[c + 1] == '\n')
				{
					c+= 2;
					nl = true;
					crlf = true;
					break;
				}
			}
			if(str[c] == '\n')
			{
				++c;
				nl = true;
				break;
			}
			++c;
		}
		nstat = ::recv(so, str, c, 0);
		if(nstat < 0)
			break;

		if(crlf)
			--nstat;

		if(nl)
			str[nstat - 1] = '\n';
	
		str += nstat;
		nleft -= nstat;
	}
	*str = 0;
	return request - (ssize_t)nleft;
}

bool Socket::isConnected(void) const
{
	return (Socket::state == SOCKET_CONNECTED) ? true : false;
}

bool Socket::isActive(void) const
{
	return (state != SOCKET_INITIAL) ? true : false;
}

bool Socket::operator!() const
{
	return (Socket::state == SOCKET_INITIAL) ? true : false;
}

void Socket::endSocket(void)
{
	if(Socket::state == SOCKET_STREAM)
	{
		state = SOCKET_INITIAL;
		if(so > -1)
		{
			close(so);
			so = -1;
		}
		return;
	}

	state = SOCKET_INITIAL;
	if(so == INVALID_SOCKET)
		return;

#ifdef	SO_LINGER
	struct linger linger;

	if(flags.linger)
	{
		linger.l_onoff = 1;
		linger.l_linger = 60;
	}
	else
		linger.l_onoff = linger.l_linger = 0;
	setsockopt(so, SOL_SOCKET, SO_LINGER, (char *)&linger, 
		(socklen_t)sizeof(linger));
#endif	
//	shutdown(so, 2);
#ifdef WIN32
	closesocket(so);
#else
	close(so);
#endif
	so = -1;
}		

#ifdef WIN32
sockerror_t Socket::connectError(void)
{
	switch(WSAGetLastError())
	{
	case WSAENETDOWN:
		return Error(SOCKET_RESOURCE_FAILURE);
	case WSAEINPROGRESS:
		return Error(SOCKET_CONNECT_BUSY);
	case WSAEADDRNOTAVAIL:
		return Error(SOCKET_CONNECT_INVALID);
	case WSAECONNREFUSED:
		return Error(SOCKET_CONNECT_REFUSED);
	case WSAENETUNREACH:
		return Error(SOCKET_CONNECT_NOROUTE);
	default:
		return Error(SOCKET_CONNECT_FAILED);
	}
}
#else
sockerror_t Socket::connectError(void)
{
	switch(errno)
	{
#ifdef	EHOSTUNREACH
	case EHOSTUNREACH:
		return Error(SOCKET_CONNECT_NOROUTE);
#endif
#ifdef	ENETUNREACH
	case ENETUNREACH:
		return Error(SOCKET_CONNECT_NOROUTE);
#endif
	case EINPROGRESS:
		return Error(SOCKET_CONNECT_BUSY);
#ifdef	EADDRNOTAVAIL
	case EADDRNOTAVAIL:
		return Error(SOCKET_CONNECT_INVALID);
#endif
	case ECONNREFUSED:
		return Error(SOCKET_CONNECT_REFUSED);
	case ETIMEDOUT:
		return Error(SOCKET_CONNECT_TIMEOUT);
	default:
		return Error(SOCKET_CONNECT_FAILED);
	}
}
#endif

sockerror_t Socket::Error(sockerror_t err, char *errs) const
{
	errid  = err;
	errstr = errs;
	if(!err)
		return err;

	if(flags.thrown)
		return err;

	// prevents recursive throws

	flags.thrown = true;
	if(getException() == THROW_OBJECT)
		throw((Socket *)this);
	return err;
}

sockerror_t Socket::setBroadcast(bool enable)
{
	int opt = (enable ? 1 : 0);
	if(setsockopt(so, SOL_SOCKET, SO_BROADCAST, 
		      (char *)&opt, (socklen_t)sizeof(opt)))
		return Error(SOCKET_BROADCAST_DENIED);
	flags.broadcast = enable;
	return SOCKET_SUCCESS;
}			

sockerror_t Socket::setKeepAlive(bool enable)
{
	int opt = (enable ? ~0: 0);
#if (defined(SO_KEEPALIVE) || defined(WIN32))
	if(setsockopt(so, SOL_SOCKET, SO_KEEPALIVE, 
		      (char *)&opt, (socklen_t)sizeof(opt)))
		return Error(SOCKET_KEEPALIVE_DENIED);
#endif
	flags.keepalive = enable;
	return SOCKET_SUCCESS;
}

sockerror_t Socket::setLinger(bool linger)
{
#ifdef	SO_LINGER
	flags.linger = linger;
	return SOCKET_SUCCESS;
#else
	return Error(SOCKET_SERVICE_UNAVAILABLE);
#endif
}

sockerror_t Socket::setRouting(bool enable)
{
	int opt = (enable ? 1 : 0);

#ifdef	SO_DONTROUTE
	if(setsockopt(so, SOL_SOCKET, SO_DONTROUTE, 
		      (char *)&opt, (socklen_t)sizeof(opt)))
		return Error(SOCKET_ROUTING_DENIED);
#endif
	flags.route = enable;
	return SOCKET_SUCCESS;
}

sockerror_t Socket::setNoDelay(bool enable)
{
	int opt = (enable ? 1 : 0);

	if(setsockopt(so, IPPROTO_TCP, TCP_NODELAY, 
		      (char *)&opt, (socklen_t)sizeof(opt)))
		return Error(SOCKET_NODELAY_ERROR);

	return SOCKET_SUCCESS;
}

sockerror_t Socket::setTypeOfService(socktos_t service)
{
#ifdef	SOL_IP
	unsigned char tos;
	switch(service)
	{
#ifdef 	IPTOS_LOWDELAY
	case SOCKET_IPTOS_LOWDELAY:
		tos = IPTOS_LOWDELAY;
		break;
#endif
#ifdef 	IPTOS_THROUGHPUT
	case SOCKET_IPTOS_THROUGHPUT:
		tos = IPTOS_THROUGHPUT;
		break;
#endif
#ifdef	IPTOS_RELIABILITY
	case SOCKET_IPTOS_RELIABILITY:
		tos = IPTOS_RELIABILITY;
		break;
#endif
#ifdef	IPTOS_MINCOST
	case SOCKET_IPTOS_MINCOST:
		tos = IPTOS_MINCOST;
		break;
#endif
	default:
		return Error(SOCKET_SERVICE_UNAVAILABLE);
	}
	if(setsockopt(so, SOL_IP, IP_TOS,(char *)&tos, (socklen_t)sizeof(tos)))
		return Error(SOCKET_SERVICE_DENIED);
	return SOCKET_SUCCESS;
#else
	return Error(SOCKET_SERVICE_UNAVAILABLE);
#endif
}

sockerror_t Socket::setTimeToLive(unsigned char ttl)
{
#ifdef	IP_MULTICAST_TTL
	if(!flags.multicast)
		return Error(SOCKET_MULTICAST_DISABLED);

	flags.ttl = ttl;
	setsockopt(so, IPPROTO_IP, IP_MULTICAST_TTL, (char *)&ttl, sizeof(ttl));
	return SOCKET_SUCCESS;
#else
	return Error(SOCKET_SERVICE_UNAVAILABLE);
#endif
}

sockerror_t Socket::setLoopback(bool enable)
{
#ifdef	IP_MULTICAST_LOOP
	unsigned char loop;

	if(!flags.multicast)
		return Error(SOCKET_MULTICAST_DISABLED);

	if(enable)
		loop = 1;
	else
		loop = 0;
	flags.loopback = enable;
	setsockopt(so, IPPROTO_IP, IP_MULTICAST_LOOP, (char *)&loop, sizeof(loop));
	return SOCKET_SUCCESS;		
#else
	return Error(SOCKET_SERVICE_UNAVAILABLE);
#endif
}	

bool Socket::isPending(sockpend_t pending, timeout_t timeout)
{
	int status;
#ifdef HAVE_POLL
	struct pollfd pfd;

	pfd.fd = so;
	pfd.revents = 0;
	switch(pending)
	{
	case SOCKET_PENDING_INPUT:
		pfd.events = POLLIN;
		break;
	case SOCKET_PENDING_OUTPUT:
		pfd.events = POLLOUT;
		break;
	case SOCKET_PENDING_ERROR:
		pfd.events = POLLERR | POLLHUP;
		break;
	}

	status = 0;
	while(status < 1)
	{

		if(timeout == TIMEOUT_INF)
			status = poll(&pfd, 1, -1);
		else
			status = poll(&pfd, 1, timeout);

		if(status < 1)
		{
			// don't stop polling because of a simple
			// signal :)
			if(status == -1 && errno == EINTR)
				continue;

			return false;
		}
	}

	if(pfd.revents & pfd.events)
		return true;
	return false;
#else
	struct timeval tv;
	struct timeval *tvp = &tv;
	fd_set grp;

	if(timeout == TIMEOUT_INF)
		tvp = NULL;
	else
	{
		tv.tv_usec = (timeout % 1000) * 1000;
		tv.tv_sec = timeout / 1000;
	}

	FD_ZERO(&grp);
	FD_SET(so, &grp);
	switch(pending)
	{
	case SOCKET_PENDING_INPUT:
		status = select(so + 1, &grp, NULL, NULL, tvp);
		break;
	case SOCKET_PENDING_OUTPUT:
		status = select(so + 1, NULL, &grp, NULL, tvp);
		break;
	case SOCKET_PENDING_ERROR:
		status = select(so + 1, NULL, NULL, &grp, tvp);
		break;
	}
	if(status < 1)
		return false;
	if(FD_ISSET(so, &grp))
		return true;
	return false;
#endif
}

Socket &Socket::operator=(const Socket &from)
{
	if(so == from.so)
		return *this;

	if(state != SOCKET_INITIAL)
		endSocket();

#ifdef WIN32
	so = from.so;
#else
	so = dup(from.so);
#endif
	if(so == INVALID_SOCKET)
	{
#ifndef WIN32
		Error(SOCKET_COPY_FAILED);
#endif
		state = SOCKET_INITIAL;
	}
	else
		state = from.state;

	return *this;
}

void Socket::setCompletion(bool immediate)
{
	flags.completion = immediate;
#ifdef WIN32
	// note that this will not work on some versions of Windows for Workgroups. Tough. -- jfc
	switch( immediate )
	{
	case false:
		// this will not work if you are using WSAAsyncSelect or WSAEventSelect.
		// -- perhaps I should throw an exception
		ioctlsocket( so, FIONBIO, (unsigned long *) 1);
		break;
	case true:
		ioctlsocket( so, FIONBIO, (unsigned long *) 0);
		break;
	}
#else
	int fflags = fcntl(so, F_GETFL);

	switch( immediate )
	{
	case false:
		fflags |= O_NONBLOCK;
		fcntl(so, F_SETFL, fflags);
		break;
	case true:
		fflags &=~ O_NONBLOCK;
		fcntl(so, F_SETFL, fflags);
		break;
	}
#endif
}

InetHostAddress Socket::getSender(tpport_t *port) const
{
	struct sockaddr_in from;
	char buf;
	socklen_t len = sizeof(from);
	int rc = ::recvfrom(so, &buf, 1, MSG_PEEK, 
			    (struct sockaddr *)&from, &len);

	if(rc == 1)
	{
		if(port)
			*port = ntohs(from.sin_port);
	}
	else
	{
		if(port)
			*port = 0;
		memset(&from.sin_addr, 0, sizeof(from.sin_addr));
	}
	if(rc < 0)
	{
#ifdef WIN32
		Error(SOCKET_INPUT_ERROR);
#else
		switch(errno)
		{
		case EINTR:
			Error(SOCKET_INPUT_INTERRUPT);
			break;
		default:
			Error(SOCKET_INPUT_ERROR);
		}
#endif
	}

	return InetHostAddress(from.sin_addr);
}

InetHostAddress Socket::getLocal(tpport_t *port) const
{
	struct sockaddr_in addr;
	socklen_t len = sizeof(addr);

	if(getsockname(so, (struct sockaddr *)&addr, &len))
	{
		Error(SOCKET_RESOURCE_FAILURE);
	      	if(port)
			*port = 0;
		memset(&addr.sin_addr, 0, sizeof(addr.sin_addr));
	}
	else
	{
		if(port)
		    	*port = ntohs(addr.sin_port);
	}
	return InetHostAddress(addr.sin_addr);
}

InetHostAddress Socket::getPeer(tpport_t *port) const
{
	struct sockaddr_in addr;
	socklen_t len = sizeof(addr);

	if(getpeername(so, (struct sockaddr *)&addr, &len))
	{
#ifndef WIN32
		if(errno == ENOTCONN)
			Error(SOCKET_NOT_CONNECTED);
		else
#endif
			Error(SOCKET_RESOURCE_FAILURE);
		if(port)
			*port = 0;
		memset(&addr.sin_addr, 0, sizeof(addr.sin_addr));
	}
	else
	{
		if(port)
	    		*port = ntohs(addr.sin_port);
	}
	return InetHostAddress(addr.sin_addr);
}

sockerror_t Socket::setMulticast(bool enable)
{
#ifdef	IP_MULTICAST_IF
	struct sockaddr_in addr;
	socklen_t len = sizeof(addr);

	if(enable == flags.multicast)
		return SOCKET_SUCCESS;

	flags.multicast = enable;
	if(enable)
		getsockname(so, (struct sockaddr *)&addr, &len);
	else
		memset(&addr.sin_addr, 0, sizeof(addr.sin_addr));
		
	setsockopt(so, IPPROTO_IP, IP_MULTICAST_IF, (char *)&addr.sin_addr, sizeof(addr.sin_addr));
	return SOCKET_SUCCESS;
#else
	return Error(SOCKET_SERVICE_UNAVAILABLE);
#endif
}

sockerror_t Socket::Join(const InetMcastAddress &ia)
{
#ifdef	IP_ADD_MEMBERSHIP
	struct ip_mreq group;
	struct sockaddr_in myaddr;
	socklen_t len = sizeof(myaddr);

	if(!flags.multicast)
		return Error(SOCKET_MULTICAST_DISABLED);

	getsockname(so, (struct sockaddr *)&myaddr, &len);
	memcpy(&group.imr_interface, &myaddr.sin_addr, sizeof(&myaddr.sin_addr));
	group.imr_multiaddr = getaddress(ia);
	setsockopt(so, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char *)&group, sizeof(group));
	return SOCKET_SUCCESS;
#else
	return Error(SOCKET_SERVICE_UNAVAILABLE);
#endif
}
 
sockerror_t Socket::Drop(const InetMcastAddress &ia)
{
#ifdef	IP_DROP_MEMBERSHIP
	struct ip_mreq group;
	struct sockaddr_in myaddr;
	socklen_t len = sizeof(myaddr);

	if(!flags.multicast)
		return Error(SOCKET_MULTICAST_DISABLED);

	getsockname(so, (struct sockaddr *)&myaddr, &len);
	memcpy(&group.imr_interface, &myaddr.sin_addr, sizeof(&myaddr.sin_addr));
	group.imr_multiaddr = getaddress(ia);
	setsockopt(so, IPPROTO_IP, IP_DROP_MEMBERSHIP, (char *)&group, sizeof(group));
	return SOCKET_SUCCESS;
#else
	return Error(SOCKET_SERVICE_UNAVAILABLE);
#endif
}

UDPSocket::UDPSocket(void) : 
Socket(AF_INET, SOCK_DGRAM, 0)
{
	memset(&peer, 0, sizeof(peer));
	peer.sin_family = AF_INET;
}

UDPSocket::UDPSocket(const InetAddress &ia, tpport_t port) : 
Socket(AF_INET, SOCK_DGRAM, 0)
{
	memset(&peer, 0, sizeof(peer));
	peer.sin_family = AF_INET;
	peer.sin_addr = getaddress(ia);
	peer.sin_port = htons(port);
#ifdef	SO_REUSEADDR
	int opt = 1;
	setsockopt(so, SOL_SOCKET, SO_REUSEADDR, (char *)&opt, (socklen_t)sizeof(opt));
#endif
	if(bind(so, (struct sockaddr *)&peer, sizeof(peer)))
	{
		endSocket();
		Error(SOCKET_BINDING_FAILED);
		return;
	}
	state = SOCKET_BOUND;
}

void UDPSocket::setPeer(const InetHostAddress &ia, tpport_t port)
{
	memset(&peer, 0, sizeof(peer));
	peer.sin_family = AF_INET;
	peer.sin_addr = getaddress(ia);
	peer.sin_port = htons(port);
}

InetHostAddress UDPSocket::getPeer(tpport_t *port) const
{
	char buf;
	socklen_t len = sizeof(peer);
	int rtn = ::recvfrom(so, &buf, 1, MSG_PEEK, (struct sockaddr *)&peer, &len);
 	
	if(rtn < 1)
	{
		if(port)
			*port = 0;
	
		memset((void*) &peer, 0, sizeof(peer));
	}
	else
	{
		if(port)
			*port = ntohs(peer.sin_port);
	}
	return InetHostAddress(peer.sin_addr);
}	

UDPBroadcast::UDPBroadcast(const InetAddress &ia, tpport_t port) : 
UDPSocket(ia, port)
{
	if(so != INVALID_SOCKET)
		setBroadcast(true);
}

void UDPBroadcast::setPeer(const BroadcastAddress &ia, tpport_t port)
{
	memset(&peer, 0, sizeof(peer));
	peer.sin_family = AF_INET;
	peer.sin_addr = getaddress(ia);
	peer.sin_port = htons(port);
}

TCPSocket::TCPSocket(const InetAddress &ia, tpport_t port, int backlog) : 
Socket(AF_INET, SOCK_STREAM, 0)
{
	struct sockaddr_in addr;

	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr = getaddress(ia);
	addr.sin_port = htons(port);

#ifdef	SO_REUSEADDR
	int opt = 1;
	setsockopt(so, SOL_SOCKET, SO_REUSEADDR, (char *)&opt, (socklen_t)sizeof(opt));
#endif
	if(bind(so, (struct sockaddr *)&addr, sizeof(addr)))
	{
		endSocket();
		Error(SOCKET_BINDING_FAILED);
		return;
	}

	if(listen(so, backlog))
	{
		endSocket();
		Error(SOCKET_BINDING_FAILED);
		return;
	}
	state = SOCKET_BOUND;
}

void TCPSocket::Reject(void)
{
	int rej = accept(so, NULL, NULL);
	shutdown(rej, 2);
#ifdef WIN32
	closesocket(rej);
#else
	close(rej);
#endif
}

TCPStream::TCPStream(TCPSocket &server, int size, bool throwflag, timeout_t to) : 
	Socket(accept(server.so, NULL, NULL))
	,streambuf() 
#ifdef	HAVE_OLD_IOSTREAM
	,iostream()
#else
	,iostream((streambuf *)this)
#endif
	,bufsize(0)
	,gbuf(NULL)
	,pbuf(NULL)
{
	tpport_t port;

#ifdef	HAVE_OLD_IOSTREAM
	init((streambuf *)this);
#endif

	timeout = to;
	setError(throwflag);
	InetHostAddress host = getPeer(&port);
	if(!server.OnAccept(host, port))
	{
		endSocket();
		Error(SOCKET_CONNECT_REJECTED);
		clear(ios::failbit | rdstate());
		return;
	}

	Allocate(size);
	Socket::state = SOCKET_CONNECTED;
}

TCPStream::TCPStream(const InetHostAddress &host, tpport_t port, int size, bool throwflag, timeout_t to) :
	Socket(AF_INET, SOCK_STREAM, 0), 
	streambuf(), 
#ifdef	HAVE_OLD_IOSTREAM
	iostream(),
#else
        iostream((streambuf *)this),
#endif
	bufsize(0),gbuf(NULL),pbuf(NULL)
{
#ifdef	HAVE_OLD_IOSTREAM
	init((streambuf *)this);
#endif
	timeout = to;
	setError(throwflag);
	Connect(host, port, size);
}

void TCPStream::Connect(const InetHostAddress &host, tpport_t port, int size)
{
	size_t i;
	for(i = 0 ; i < host.getAddressCount(); i++)
	{
		struct sockaddr_in addr;
		memset(&addr, 0, sizeof(addr));
		addr.sin_family = AF_INET;
		addr.sin_addr = host.getAddress(i);
		addr.sin_port = htons(port);
		
		if(connect(so, (struct sockaddr *)&addr, (socklen_t)sizeof(addr)) == 0)
			break;
	}

	if(i == host.getAddressCount())
	{
		connectError();
		endSocket();
		return;
	}

	Allocate(size);
	Socket::state = SOCKET_CONNECTED;
}

TCPStream::TCPStream(bool throwflag) :
	Socket(PF_INET, SOCK_STREAM, 0),
	streambuf(), 
#ifdef	HAVE_OLD_IOSTREAM
	iostream(),
#else
        iostream((streambuf *)this), 
#endif
        bufsize(0),gbuf(NULL),pbuf(NULL),
        timeout(0)
{ 
#ifdef	HAVE_OLD_IOSTREAM
	init((streambuf *)this);
#endif
	setError(throwflag);
}	

TCPStream::TCPStream(const TCPStream &source) :
Socket(dup(source.so)), streambuf(), 
#ifdef	HAVE_OLD_IOSTREAM
iostream()
#else
iostream((streambuf *)this) 
#endif
{
#ifdef	HAVE_OLD_IOSTREAM
	init((streambuf *)this);
#endif
	bufsize = source.bufsize;
	Allocate(bufsize);
}

void TCPStream::endStream(void) 
{
	if(bufsize) sync();
	delete [] gbuf;
	delete [] pbuf;
	gbuf = pbuf = NULL;
	bufsize = 0;
	endSocket();
}

void TCPStream::Allocate(int size)
{
	if(size < 2)
	{
		bufsize = 1;
		return;
	}

	gbuf = new char[size];
	pbuf = new char[size];
	if(!pbuf || !gbuf)
	{
		Error(SOCKET_RESOURCE_FAILURE);
		return;
	}
	bufsize = size;
	clear();

#if (__GNUC__ < 3) && !defined(WIN32)
	setb(gbuf, gbuf + size, 0);
#endif
	setg(gbuf, gbuf + size, gbuf + size);
	setp(pbuf, pbuf + size);
}

int TCPStream::doallocate()
{
	if(bufsize)
		return 0;

	Allocate(1);
	return 1;
}

int TCPStream::uflow(void)
{
	int ret = underflow(); 

	if (ret == EOF)
		return EOF; 

	if (bufsize != 1)
		gbump(1); 

	return ret; 
}

int TCPStream::underflow(void)
{
	int rlen = 1;
	unsigned char ch;

	if(bufsize == 1)
	{
		if(Socket::state == SOCKET_STREAM)
			rlen = ::read(so, (char *)&ch, 1);
		else if(timeout && !Socket::isPending(SOCKET_PENDING_INPUT, timeout))
        {
            clear(ios::failbit | rdstate());
            Error(SOCKET_TIMEOUT_ERROR);
            return EOF;
		}
		else
			rlen = ::recv(so, (char *)&ch, 1, 0);
		if(rlen < 1)
		{
			if(rlen < 0)
			{
				clear(ios::failbit | rdstate());
				Error(SOCKET_INPUT_ERROR);
			}
			return EOF;
		}
		return ch;
	}

	if(!gptr())
		return EOF;

	if(gptr() < egptr())
		return (unsigned char)*gptr();

	rlen = (gbuf + bufsize) - eback();
	if(Socket::state == SOCKET_STREAM)
		rlen = ::read(so, (char *)eback(), rlen);
	else if(timeout && !Socket::isPending(SOCKET_PENDING_INPUT, timeout))
	{
		clear(ios::failbit | rdstate());
		Error(SOCKET_TIMEOUT_ERROR);
		return EOF;
	}
	else
		rlen = ::recv(so, (char *)eback(), rlen, 0);
	if(rlen < 1)
	{
		if(rlen < 0)
		{
			clear(ios::failbit | rdstate());
			Error(SOCKET_INPUT_ERROR);
		}
		return EOF;
	}

	setg(eback(), eback(), eback() + rlen);
	return (unsigned char) *gptr();
}

bool TCPStream::isPending(sockpend_t pending, timeout_t timeout)
{
	if(pending == SOCKET_PENDING_INPUT && in_avail())
		return true;
	else if(pending == SOCKET_PENDING_OUTPUT)
		flush();

	return Socket::isPending(pending, timeout);
}

int TCPStream::sync(void)
{
	overflow(EOF);
	setg(gbuf, gbuf + bufsize, gbuf + bufsize);
	return 0;
}

int TCPStream::overflow(int c)
{
	unsigned char ch;
	int rlen, req;

	if(bufsize == 1)
	{
		if(c == EOF)
			return 0;

		ch = (unsigned char)(c);
		if(Socket::state == SOCKET_STREAM)
			rlen = ::write(so, (const char *)&ch, 1);
		else
			rlen = ::send(so, (const char *)&ch, 1, 0);
		if(rlen < 1)
		{
			if(rlen < 0)
			{
				clear(ios::failbit | rdstate());
				Error(SOCKET_OUTPUT_ERROR);
			}
			return EOF;
		}
		else
			return c;
	}		

	if(!pbase())
		return EOF;

	req = pptr() - pbase();
	if(req)
	{
		if(Socket::state == SOCKET_STREAM)
			rlen = ::write(so, (const char *)pbase(), req);
		else
			rlen = ::send(so, (const char *)pbase(), req, 0);
		if(rlen < 1)
		{
			if(rlen < 0)
			{
				clear(ios::failbit | rdstate());
				Error(SOCKET_OUTPUT_ERROR);
			}
			return EOF;
		}
		req -= rlen;
	}

	// if write "partial", rebuffer remainder
	
	if(req)
		memcpy(pptr(), pptr() + rlen, req);
	setp(pbuf + req, pbuf + bufsize);

	if(c != EOF)
	{
		*pptr() = (unsigned char)c;
		pbump(1);
	}
	return c;
}

tcpstream::tcpstream() : 
TCPStream()
{
	setError(false);	/* no exceptions */
}

tcpstream::tcpstream(const char *addr, int buf) : 
TCPStream()
{
	setError(false);
	open(addr, buf);
}

tcpstream::tcpstream(TCPSocket &server, int buf) :
TCPStream()
{
	setError(false);
	open(server, buf);
}

bool tcpstream::operator!() const
{
	return (Socket::state != SOCKET_CONNECTED) ? true : false;
}

void tcpstream::open(TCPSocket &tcp, int buf)
{
	tpport_t port;
	InetHostAddress host;

	endStream();
	so = accept(tcp.so, NULL, NULL);
	if(so == INVALID_SOCKET)
		return;

	host = getPeer(&port);
	if(!tcp.OnAccept(host, port))
	{
		endSocket();
		clear(ios::failbit | rdstate());
		return;
	}

	Allocate(buf);
	Socket::state = SOCKET_CONNECTED;
}
	
void tcpstream::open(const char *path, int buf)
{
	char abuf[256];
	const char *cp;
	char *ep;
	InetHostAddress host;

	close();
	if(Socket::state != SOCKET_AVAILABLE)
		return;
	
	cp = strrchr(path, ':');
	if(!cp)
		return;

	++cp;
	strncpy(abuf, path, sizeof(abuf) - 1);
	abuf[sizeof(abuf) - 1] = 0;
	ep = strrchr(abuf, ':');
	if(ep)
		*ep = 0;

	host = abuf;

	Connect(host, (tpport_t)atoi(cp), buf);
}
	
void tcpstream::close(void)
{
	if(Socket::state == SOCKET_AVAILABLE)
		return;

	endStream();
	so = socket(AF_INET, SOCK_STREAM, 0);
	if(so != INVALID_SOCKET)
		Socket::state = SOCKET_AVAILABLE;
}

TCPSession::TCPSession(Semaphore *start, const InetHostAddress &ia, 
		       tpport_t port, int size, int pri, int stack) :
TCPStream(), Thread(start, pri, stack)
{
	setCompletion(false);
	setError(false);
	Allocate(size);
	
	size_t i;
	for(i = 0 ; i < ia.getAddressCount(); i++)
	{
		struct sockaddr_in addr;
		memset(&addr, 0, sizeof(addr));
		addr.sin_family = AF_INET;
		addr.sin_addr = ia.getAddress(i);
		addr.sin_port = htons(port);
		
		if(connect(so, (struct sockaddr *)&addr, (socklen_t)sizeof(addr)) == 0)
			break;
		
#ifdef WIN32
//		if(WSAGetLastError() == WSAEWOULDBLOCK)
		if(WSAGetLastError() == WSAEISCONN)
#else
		if(errno == EINPROGRESS)
#endif
		{
			Socket::state = SOCKET_CONNECTING;
			return;
		}
	}

	if(i == ia.getAddressCount())
	{
		endSocket();
		Socket::state = SOCKET_INITIAL;
		return;
	}

	setCompletion(true);
	Socket::state = SOCKET_CONNECTED;
}
	

TCPSession::TCPSession(Semaphore *start, TCPSocket &s, 
		       int size, int pri, int stack) :
TCPStream(s, size), Thread(start, pri, stack)
{
	setCompletion(true);
	setError(false);
}

int TCPSession::WaitConnection(timeout_t timeout)
{
	long sockopt = 0;
	socklen_t len = sizeof(sockopt);

	switch(Socket::state)
	{
	case SOCKET_INITIAL:
		return -1;
	case SOCKET_CONNECTED:
		break;
	case SOCKET_CONNECTING:	
		if(!Socket::isPending(SOCKET_PENDING_OUTPUT, timeout))
		{
			endSocket();
			Socket::state = SOCKET_INITIAL;
			return -1;
		}
	
		getsockopt(so, SOL_SOCKET, SO_ERROR, (char *)&sockopt, &len);
		if(sockopt)
		{
			endSocket();
			Socket::state = SOCKET_INITIAL;
			return -1;
		}
	}
	Socket::state = SOCKET_CONNECTED;
	return 0;	
}

void TCPSession::Initial(void)
{
	if(WaitConnection(60000))
		Exit();
}

ostream &operator<<(ostream &os, const InetAddress &ia)
{
	os << inet_ntoa(getaddress(ia));
	return os;
}

#ifdef WIN32
init_WSA::init_WSA()
{
	//-initialize OS socket resources!

	if (WSAStartup(MAKEWORD(2, 2), &wsaData))
	{
		abort();
	}
};

init_WSA init_wsa;
#endif

/** EMACS **
 * Local variables:
 * mode: c++
 * c-basic-offset: 8
 * End:
 */
